#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
import urllib

from django.utils.functional import cached_property
from opentracing import Format
from opentracing.ext import tags
from .conf import *
from .request_context import span_in_context, span_out_context, get_current_span

logger = logging.getLogger(__name__)


class TraceMiddleware:
    """"use jaeger_client realizing tracing"""

    def __init__(self, get_response=None):
        self.get_response = get_response

        # Initialize tracer
        self.tracer = self._tracer

    @cached_property
    def _tracer(self):
        from .initial_tracer import initialize_global_tracer
        return initialize_global_tracer()

    def __call__(self, request):
        if self._ignore_request(request):
            return self.get_response(request)

        self._parse_wsgi_headers(request)
        self.full_url(request)

        try:
            parent_ctx = self.tracer.extract(
                Format.HTTP_HEADERS,
                carrier=request.headers
            )
        except Exception as e:
            logger.exception(f'Failed to extract parent context:{e}')
            parent_ctx = None

        span = self._gen_span(request, parent_ctx)
        span_in_context(span)

        # # Add tracing headers to request
        carrier = {}
        try:
            self._tracer.inject(
                span_context=span.context,
                format=Format.HTTP_HEADERS,
                carrier=carrier
            )
            for key, value in carrier.items():
                request.headers[key] = value
        except Exception as e:
            logger.debug(f"Failed to inject tracing headers: {e}")

        response = self.get_response(request)
        trace_id_header = get_tracer_config().get("trace_id_header", "trace-id")
        response[trace_id_header] = span.trace_id

        try:
            span.set_tag(tags.HTTP_STATUS_CODE, response.status_code)
        except Exception as e:
            logger.exception(f'Error setting response tags for tracing:{e}')
        finally:
            span.finish()
            span_out_context()

        return response

    @staticmethod
    def _ignore_request(request) -> bool:
        """
        Check if the request should be ignored based on configuration.
        :param: request:
        :return: True if request should be ignored, False otherwise.
        """
        ignore_urls = get_tracing_config().get("http_requests", {}).get("ignore_urls", [])
        request_path = request.path_info

        return any(
            ignore_url in request_path
            for ignore_url in ignore_urls
        )

    @staticmethod
    def _parse_wsgi_headers(request):
        """
        HTTP headers are presented in WSGI environment with 'HTTP_' prefix.
        This method finds those headers, removes the prefix, converts
        underscores to dashes, and converts to lower case.
        :param request:
        :return: returns a dictionary of headers
        """
        prefix = 'HTTP_'
        p_len = len(prefix)
        # use .items() despite suspected memory pressure bc GC occasionally
        # collects wsgi_environ.iteritems() during iteration.
        headers = {
            key[p_len:].replace('_', '-').lower():
                val for (key, val) in request.environ.items()
            if key.startswith(prefix)}
        setattr(request, 'headers', headers)

    @staticmethod
    def full_url(request):
        """
        Build the full URL from WSGI environ variables.
        Taken from:
        http://legacy.python.org/dev/peps/pep-3333/#url-reconstruction
        :return: Reconstructed URL from WSGI environment.
        """
        environ = request.environ
        url = environ['wsgi.url_scheme'] + '://'

        if environ.get('HTTP_HOST'):
            url += environ['HTTP_HOST']
        else:
            url += environ['SERVER_NAME']

            if environ['wsgi.url_scheme'] == 'https':
                if environ['SERVER_PORT'] != '443':
                    url += ':' + environ['SERVER_PORT']
            else:
                if environ['SERVER_PORT'] != '80':
                    url += ':' + environ['SERVER_PORT']

        url += urllib.parse.quote(environ.get('SCRIPT_NAME', ''))
        url += urllib.parse.quote(environ.get('PATH_INFO', ''))
        if environ.get('QUERY_STRING'):
            url += '?' + environ['QUERY_STRING']
        setattr(request, 'full_url', url)

    def _gen_span(self, request, parent_ctx):
        """
        Create a span for the request.
        :param request:
        :param parent_ctx:
        :return:
        """
        operation_name = '{} {}'.format(request.method, request.path)

        span_tags = {
            tags.SPAN_KIND: tags.SPAN_KIND_RPC_SERVER,
            tags.HTTP_URL: request.full_url,
            tags.HTTP_METHOD: request.method,
            tags.COMPONENT: get_service_name()
        }

        remote_ip = request.environ.get('REMOTE_ADDR')
        if remote_ip:
            span_tags[tags.PEER_HOST_IPV4] = remote_ip

        remote_port = request.environ.get('REMOTE_PORT')
        if remote_port:
            span_tags[tags.PEER_PORT] = remote_port

        user_agent = request.META.get("HTTP_USER_AGENT")
        if user_agent:
            max_length = get_tracing_config().get("http_requests", {}).get("max_tag_value_length", 1024)
            span_tags["http.user_agent"] = user_agent[:max_length]

        span = self._tracer.start_span(
            operation_name=operation_name,
            child_of=parent_ctx,
            tags=span_tags)

        return span


def _tracing_injection(func):
    def _call(*args, **kwargs):
        request = args[1]

        try:
            span = get_current_span()
            if span:
                trace_id_header = get_tracer_config().get("trace_id_header", "trace-id")
                request.headers[trace_id_header] = span.trace_id
        except Exception:
            pass
        return func(*args, **kwargs)  # actual call

    return _call


sys = __import__("sys")
session = sys.modules['requests.sessions']
session.Session.prepare_request = _tracing_injection(session.Session.prepare_request)



"""
Database instrumentation for Django ORM queries.
Automatically traces database operations with query details and performance metrics.
"""
import logging

from opentracing.ext import tags

from ..conf import is_component_enabled, get_tracing_config, get_service_name
from .utils import tracing_enabled

logger = logging.getLogger(__name__)


def db_name(func, *args, **kwargs):
    return "DB_QUERY"


def db_span_processor(func, span, *args, **kwargs):
    span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
    span.set_tag(tags.COMPONENT, "django.db")
    span.set_tag(tags.DATABASE_TYPE, "sql")
    span.set_tag(tags.DATABASE_INSTANCE, get_service_name())

    # Add SQL statement (optionally truncated)
    max_length = get_tracing_config().get("database", {}).get("max_query_length", 1000)
    span.set_tag(tags.DATABASE_STATEMENT, args[0][:max_length])


def db_need_ignore(func, *args, **kwargs):
    sql = args[0]  # SQL query
    ignore_sqls = get_tracing_config().get("database", {}).get("ignore_sqls", [])
    return any(
        ignore_sql.upper() in sql.upper()
        for ignore_sql in ignore_sqls
    )


class DatabaseInstrumentation:
    """Database instrumentation manager."""

    @classmethod
    def install(cls):
        """Install database instrumentation."""
        if not is_component_enabled("database"):
            return

        from django.db.backends.utils import CursorWrapper

        CursorWrapper.execute = tracing_enabled(CursorWrapper.execute, db_name, db_span_processor, db_need_ignore)
        logger.info("Database instrumentation installed")



"""
Redis instrumentation for Redis operations.
Automatically traces Redis commands with performance metrics.
"""
import logging

from opentracing import logs
from opentracing.ext import tags

from .utils import tracing_enabled
from ..conf import is_component_enabled, get_tracing_config

logger = logging.getLogger(__name__)


def redis_name(func, *args, **kwargs):
    return "REDIS"


def redis_span_processor(func, span, *args, **kwargs):
    span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
    span.set_tag(tags.COMPONENT, "redis")
    span.set_tag(tags.DATABASE_TYPE, "redis")

    # Add SQL statement (optionally truncated)
    max_length = get_tracing_config().get("database", {}).get("max_query_length", 1000)
    statement = args[1] + " " + args[2]
    span.set_tag(logs.MESSAGE, statement[:max_length])
    span.set_tag(tags.DATABASE_STATEMENT, statement[:max_length])


def redis_need_ignore(func, *args, **kwargs):
    ignore_commands = get_tracing_config().get("redis", {}).get("ignore_commands", [])
    return any(
        ignore_command.upper() in args[0].upper()
        for ignore_command in ignore_commands
    )


class RedisInstrumentation:
    """Redis instrumentation manager."""

    @classmethod
    def install(cls):
        """Install Redis instrumentation."""
        if not is_component_enabled("redis"):
            return

        try:
            import redis
            from redis.client import Redis

            Redis.execute_command = tracing_enabled(Redis.execute_command, redis_name, redis_span_processor, redis_need_ignore)
            logger.info("Redis instrumentation installed")
        except ImportError:
            logger.warning("Redis package not found, skipping Redis instrumentation")



"""
HTTP client instrumentation for outgoing requests.
Automatically traces HTTP requests made using the requests library.
"""
import logging
import requests

from opentracing import Format
from opentracing.ext import tags
from requests.adapters import HTTPAdapter
from urllib3.util import parse_url

from ..conf import is_component_enabled, get_tracing_config
from ..initial_tracer import initialize_global_tracer
from ..request_context import get_current_span

logger = logging.getLogger(__name__)


class TracingHTTPAdapter(HTTPAdapter):
    """HTTP adapter that adds tracing to requests."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tracer = initialize_global_tracer()
        self._config = get_tracing_config().get("http_requests", {})

    def _should_ignore_tracing(self, request) -> bool:
        """Check if request should be traced."""
        ignore_urls = self._config.get("ignore_urls", [])
        request_url = request.url

        return any(
            ignore_url in request_url
            for ignore_url in ignore_urls
        )

    def send(self, request, **kwargs):
        """Send request with tracing."""
        if self._should_ignore_tracing(request):
            return super().send(request, **kwargs)

        span = self._create_span(request)

        try:
            self._inject_headers(request, span)
            response = super().send(request, **kwargs)

            # Add response information to span
            if response:
                span.set_tag(tags.HTTP_STATUS_CODE, response.status_code)

                if response.status_code >= 400:
                    span.set_tag(tags.ERROR, True)

            return response

        except Exception as e:
            span.set_tag(tags.ERROR, True)
            span.log_kv({
                "event": "error",
                "error.kind": e.__class__.__name__,
                "error.object": str(e),
                "message": str(e),
            })
            raise
        finally:
            span.finish()

    def _create_span(self, request):
        """Create tracing span for HTTP request."""
        parent_span = get_current_span()
        parsed_url = parse_url(request.url)

        operation_name = f"{request.method} {parsed_url.path or '/'}"

        span = self._tracer.start_span(
            operation_name=operation_name,
            child_of=parent_span
        )

        # Set standard HTTP tags
        span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
        span.set_tag(tags.HTTP_URL, request.url)
        span.set_tag(tags.HTTP_METHOD, request.method)
        span.set_tag(tags.COMPONENT, "requests")

        if parsed_url.host:
            span.set_tag(tags.PEER_HOST_IPV4, parsed_url.host)
        if parsed_url.port:
            span.set_tag(tags.PEER_PORT, parsed_url.port)

        return span

    def _inject_headers(self, request, span):
        """Inject tracing headers into request."""
        carrier = {}
        try:
            self._tracer.inject(
                span_context=span.context,
                format=Format.HTTP_HEADERS,
                carrier=carrier
            )

            # Add tracing headers to request
            for key, value in carrier.items():
                request.headers[key] = value

        except Exception as e:
            logger.debug(f"Failed to inject tracing headers: {e}")



"""
Celery instrumentation for distributed task tracing.
Automatically traces Celery task execution and message passing.
"""
import logging

from opentracing.ext import tags

from ..conf import is_component_enabled
from .utils import tracing_enabled

logger = logging.getLogger(__name__)


def celery_name(func, *args, **kwargs):
    return "CELERY"


def celery_span_processor(func, span, *args, **kwargs):
    span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
    span.set_tag(tags.COMPONENT, "celery")
    span.set_tag('task.name', getattr(args[0], "name", ""))


class CeleryInstrumentation:
    """Celery instrumentation manager."""

    @classmethod
    def install(cls):
        """Install Celery instrumentation."""
        if not is_component_enabled("celery"):
            return

        try:
            from celery.app.task import Task
            Task.apply_async = tracing_enabled(Task.apply_async, celery_name, celery_span_processor)
            logger.info("Celery instrumentation installed")
        except ImportError:
            logger.warning("Celery package not found, skipping Celery instrumentation")



"""
Django Distributed Tracing Middleware

A comprehensive middleware for Django applications to implement distributed tracing
with Jaeger, supporting HTTP requests, database queries, Redis operations,
Celery tasks, and RocketMQ messaging.
"""

__version__ = "1.0.0"
__author__ = "zhaishuaishuai"
__email__ = "zhaishuaishuai001@gmail.com"

from .middleware import TraceMiddleware
from .initial_tracer import initialize_global_tracer
from .request_context import get_current_span, span_in_context, span_out_context
from .exceptions import TracingConfigError, TracingError

__all__ = [
    "TraceMiddleware",
    "initialize_global_tracer",
    "get_current_span",
    "span_in_context",
    "span_out_context",
    "TracingConfigError",
    "TracingError",
]



import opentracing

from ..request_context import get_current_span


def gen_name(func, *args, **kwargs):
    return func.__name__


def no_op(func, span, *args, **kwargs):
    pass


def no_ignore(func, *args, **kwargs) -> bool:
    return False


def tracing_enabled(func, operation_name=gen_name, span_processor=no_op, need_ignore=no_ignore):
    def call(*args, **kwargs):

        # ignore this function
        if need_ignore(func, *args, **kwargs):
            return func(*args, **kwargs)

        parent_span = get_current_span()
        tracer = opentracing.global_tracer()

        with tracer.start_span(operation_name(func, *args, **kwargs), child_of=parent_span) as span:
            span_processor(func, span, *args, **kwargs)
            try:
                return func(*args, **kwargs)
            except Exception as e:
                span.set_tag('error', True)
                span.log_kv({'event': 'error', 'error.object': e})
                raise

    return call



"""
Custom exceptions for the tracing package.
"""


class TracingError(Exception):
    """Base exception for tracing-related errors."""
    pass


class TracingConfigError(TracingError):
    """Exception raised for configuration errors."""
    pass


class TracingInitializationError(TracingError):
    """Exception raised when tracer initialization fails."""
    pass


class SpanNotFoundError(TracingError):
    """Exception raised when trying to access a span that doesn't exist."""
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

__all__ = [
    "get_tracer_config",
    "get_tracing_config",
    "get_service_name",
    "is_component_enabled",
]


# Default tracer configuration for Jaeger
DEFAULT_TRACER_CONFIG = {
    'sampler': {
        'type': 'const',
        'param': 1,
    },
    'local_agent': {
        'reporting_host': 'localhost',
        'reporting_port': 6831,
    },
    'trace_id_header': 'trace-id',
    'baggage_header_prefix': 'jaegertrace-',
    "logging": True,
}

# Default tracing configuration for different components
DEFAULT_TRACING_CONFIG = {
    "http_requests": {
        "ignore_urls": [],
        "max_tag_value_length": 1024,
    },
    "database": {
        "enabled": True,
        "ignore_sqls": [],
        "max_query_length": 1000,
    },
    "redis": {
        "enabled": True,
        "ignore_commands": [],
        "max_command_length": 500,
    },
    "celery": {
        "enabled": True,
    },
}


def get_tracer_config() -> dict:
    """Get tracer configuration from Django settings with defaults."""
    config = getattr(settings, "TRACER_CONFIG", {})
    # Merge with defaults
    merged_config = DEFAULT_TRACER_CONFIG.copy()
    merged_config.update(config)
    return merged_config


def get_tracing_config() -> dict:
    """Get tracing configuration from Django settings with defaults."""
    config = getattr(settings, "TRACING_CONFIG", {})
    # Deep merge with defaults
    merged_config = {}
    for key, default_value in DEFAULT_TRACING_CONFIG.items():
        if key in config:
            if isinstance(default_value, dict):
                merged_config[key] = {**default_value, **config[key]}
            else:
                merged_config[key] = config[key]
        else:
            merged_config[key] = default_value
    return merged_config


def get_service_name() -> str:
    """Get service name for tracing."""
    service_name = getattr(settings, "TRACING_SERVICE_NAME", None)
    if not service_name:
        # Try to get from other common settings
        wsgi_application = getattr(settings, "WSGI_APPLICATION", None)
        if not wsgi_application:
            raise ImproperlyConfigured(
                "TRACING_SERVICE_NAME or WSGI_APPLICATION must be set in Django settings"
            )
        service_name = wsgi_application.split(".")[0]
    return service_name


def is_component_enabled(component_name) -> bool:
    """Check if a specific tracing component is enabled."""
    config = get_tracing_config()
    component_config = config.get(component_name, {})
    return component_config.get("enabled", False)



from django.apps import AppConfig
from django.conf import settings
from .instrumentation import *
from jaegertrace.conf import is_component_enabled


class LoggerConfig(AppConfig):
    name = 'jaegertrace'

    def ready(self):
        if not hasattr(settings, 'TRACING_CONFIG'):
            return
        if is_component_enabled('database'):
            DatabaseInstrumentation.install()
        if is_component_enabled('redis'):
            RedisInstrumentation.install()
        if is_component_enabled('celery'):
            HTTPInstrumentation.install()
